/****************************************************************************

    This file is part of The Newtonator soft synth.

    Copyright (C) 2010-2011 Michael Bechard <mbechard@users.sourceforge.net>

    The Newtonator is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The Newtonator is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Newtonator.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************/

#include "VariSource.h"
#include "config.h"
#include "INewtSettings.h"
#include <math.h>

using namespace stk;

VariSource::VariSource(INewtSettings *settings)
{
    _type = SINE;
    _amplitude = 0.0;
    _usingEnvelope = true;
    _generator = GenPtr(new SineWave());
    _set = settings;
}

VariSource::~VariSource()
{
}

void VariSource::copy(const VariSource& source)
{
    setType(source.getType());
    setFrequency(source.getFrequency());
}

StkFrames& VariSource::tick(StkFrames& frames, unsigned int channel)
{
    #ifdef INCLUDE_EXT_INPUTS
    float* extBuff = (_use == FLOOR ? _set->getFloorExtInput(channel) : _set->getGModExtInput(channel));

    if (_type == EXTERNAL)
    {
        for (size_t i = 0; i < frames.size(); i++)
        {
            frames[i] = std::max(std::min(extBuff[i], 1.0f), -1.0f);
        }
    }
    else
    #endif
    {
        frames = _generator->tick(frames, channel);
    }

    if (_usingEnvelope)
    {
        StkFrames envFrames(frames.size(), frames.channels());
        envFrames = _envelope.tick(envFrames, channel);

        for (size_t frame = channel; frame < frames.size(); frame += frames.channels())
        {
            frames[frame] = frames[frame] * envFrames[frame] * _amplitude;
        }
    }

    return frames;
}

StkFloat VariSource::tick(unsigned int frameIdx, unsigned int channel)
{
    StkFrames buffer(1, 1);
    #ifdef INCLUDE_EXT_INPUTS
    float* extBuff = (_use == FLOOR ? _set->getFloorExtInput(channel) : _set->getGModExtInput(channel));

    if (_type == EXTERNAL)
    {
        buffer[0] = std::max(std::min(extBuff[frameIdx], 1.0f), -1.0f);
    }
    else
    #endif
        _generator->tick(buffer, 0);

    return buffer[0] * (_usingEnvelope ? _envelope.tick() * _amplitude : 1.0);
}

void VariSource::setType(Type type)
{
    if (type != _type)
    {
        _type = type;

        switch (_type)
        {
            case SINE:
                _generator = GenPtr(new SineWave());
                break;
            case SAW:
                _generator = GenPtr(new BlitSaw());
                break;
            case SQUARE:
                _generator = GenPtr(new BlitSquare());
                break;
            case IMPULSE:
                _generator = GenPtr(new Blit());
                break;
            case NOISE:
                _generator = GenPtr(new Noise());
                break;
            default:
                _generator = GenPtr();
                break;
        }
    }
}

VariSource::Type VariSource::getTypeFromFloat(stk::StkFloat val)
{
    switch ((int)val)
    {
        case 0:
            return SINE;
        case 1:
            return SAW;
        case 2:
            return SQUARE;
        case 3:
            return IMPULSE;
        case 4:
            return NOISE;
        #ifdef INCLUDE_EXT_INPUTS
        case 5:
            return EXTERNAL;
        #endif
        default:
            return UNDEFINED;
    }

    return UNDEFINED;
}

void VariSource::setFrequency(StkFloat frequency, bool reset)
{
    _frequency = frequency;

    if (reset)
        this->reset();

    switch (_type)
    {
        case SINE:
            ((SineWave*)_generator.get())->setFrequency(frequency);
            break;
        case SAW:
            ((BlitSaw*)_generator.get())->setFrequency(frequency);
            break;
        case SQUARE:
            ((BlitSquare*)_generator.get())->setFrequency(frequency);
            break;
        case IMPULSE:
            ((Blit*)_generator.get())->setFrequency(frequency);
            break;
        default:
            //Don't worry about others, don't have setable frequency
            break;
    }
}

void VariSource::setHarmonics(int numHarmonics)
{
    switch (_type)
    {
        case SAW:
            ((BlitSaw*)_generator.get())->setHarmonics(numHarmonics);
            break;
        case SQUARE:
            ((BlitSquare*)_generator.get())->setHarmonics(numHarmonics);
            break;
        case IMPULSE:
            ((Blit*)_generator.get())->setHarmonics(numHarmonics);
            break;
        default:
            break;
    }
}

void VariSource::reset()
{
    switch (_type)
    {
        case SINE:
            ((SineWave*)_generator.get())->reset();
            break;
        case SAW:
            ((BlitSaw*)_generator.get())->reset();
            break;
        case SQUARE:
            ((BlitSquare*)_generator.get())->reset();
            break;
        case IMPULSE:
            ((Blit*)_generator.get())->reset();
            break;
        default:
            //Don't worry about others, don't have reset function
            break;
    }
}
